"""
##############
``peltak run``
##############


Dev Docs
========

What need to be done?
~~~~~~~~~~~~~~~~~~~~~

.. uml::

    @startwbs
    * Scripts support
    ** Render command template
    *** Config values
    *** Env variables
    *** command line options
    ** click integration
    *** Display about command as help
    *** Autocomplete script names
    @endwbs


Configuration
~~~~~~~~~~~~~

Options for a single script

.. code-block:: yaml

    scripts:
      checks:
        about: Run checks on the code base
        accepts_files: True
        command: |
          mypy {{files}}; \\
          pycodestyle --config ops/tools/pep8.ini {{files}}; \\
          pylint --rc-file ops/tools/pylint.ini {{files}};




Example
-------

.. code-block:: yaml

    scripts:
      build-docs:
        about: Generate sphinx documentation
        command: |
          sphinx-refdoc  \\
            -i src/peltak \\
            -i src/peltak_appengine \\
            -i src/peltak_django \\
            -i src/cliform \\
            {{opts.verbosity | count_flag('-v')}} \\
            docs/ref; \\
            \\
          sphinx-build \\
            -b html \\
            -d {{cfg.build_dir}}/docs \\
            docs \\
            docs/html

      unit-tests:
        about: Run unit tests
        command: |
          pytest -c ops/tools/pytest.ini \\
            --cov-config=ops/tools/coverage.ini \\
            --cov={{cfg.src_dir}} \\
            --cov-report=term \\
            --cov-report=html:.build/coverage \\
            --junitxml={cfg.build_dir}/test-results.xml \\
            --full-trace \\
            -vvv \\
            test/unit

      checks:
        about: Run checks on the code base
        command: |
          mypy {{files}}; \\
          pycodestyle --config ops/tools/pep8.ini {{files}}; \\
          pylint --rc-file ops/tools/pylint.ini {{files}};
"""
from pathlib import Path

from peltak.commands import root_cli
from peltak.core import conf, hooks
from peltak.core.scripts import loader


@root_cli.group('run')
def run_cli():
    """ Run custom scripts """
    pass  # nocov


@hooks.register('post-conf-load')
def post_conf_load():
    """ After the config was loaded, register all scripts as click commands. """
    scripts_dir = Path(conf.get_path('scripts_dir', 'scripts'))

    loader.register_scripts_from(scripts_dir)



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.hooks
    :synopsis: Runtime hooks framework.

###########
Using Hooks
###########

Hook allow responding to various events during command execution. peltak exposes
a set of standard hooks that you can use to react to what's happening. Commands
(including ones in external packages) can provide their own hooks to be used
by other commands.

To call a hook you don't have to register it, just know it's name. You can do
it simply with:

.. code-block:: python

    from peltak.commands import root_cli
    from peltak.core import hooks

    @root_cli.command('test-command')
    def test_command():
        hooks.call('pre-test-command', 3.14159, msg='hello, world!')


Registering a handler for any hook is also very simple:

.. code-block:: python

    from peltak.core import hooks

    @hooks.register('pre-test-command')
    def pre_test_command(pi, msg='<msg missing>'):
        print("pre_test_command called with ({}, {})".format(pi, msg))

"""
from typing import Any, Callable


AnyFunc = Callable[..., Any]
Decorator = Callable[[AnyFunc], AnyFunc]


class HooksRegister(object):
    """ Intermediary between hooks emitters and handlers. """
    def __init__(self):
        self.hooks = {}

    def remove(self, name: str, fn: AnyFunc):
        """ Remove a handler for a given hook.

        After this call, the *fn* function won't be called when the hook is
        called.

        Args:
            name:
                The name of the hook.
            fn:
                The function that we want to unregister from the hook calls.
        """
        handlers = self.hooks.get(name, [])
        handlers.remove(fn)

    def call(self, name: str, *args: Any, **kw: Any):
        """ Call all hooks registered for the given name.

        This will pass all arguments except for the name directly to the hook.
        Remember that hook handler should not modify those arguments as that
        will affect all hooks called later in the sequence.

        Args:
            name:
                Then name of the hook.
            *args, **kw:
                Arguments and keyword arguments that should be passed to the
                hook handlers.
        """
        if not name:
            raise ValueError("name cannot be empty")

        handlers = self.hooks.get(name, [])
        for handler_fn in handlers:
            handler_fn(*args, **kw)

    def __call__(self, name: str) -> Decorator:
        """ Return a decorator that will register the wrapped function under name.

        Args:
            name (str):
                The name of the hook you are subscribing to.
        """
        if not name:
            raise ValueError("name cannot be empty")

        def decorator(fn: AnyFunc) -> AnyFunc:  # pylint: disable=missing-docstring
            self._register_handler(name, fn)
            return fn
        return decorator

    def _register_handler(self, name: str, fn: AnyFunc) -> None:
        handlers = self.hooks.setdefault(name, [])

        if fn in handlers:
            raise ValueError("{} already registered for {}", fn, name)

        handlers.append(fn)


register = HooksRegister()



# Copyright 2017-2021 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dataclasses
from collections import namedtuple
from typing import List, Optional

from .. import conf, shell


Author = namedtuple('Author', 'name email')


@dataclasses.dataclass
class BranchDetails:
    """ Branch name parsed into type and title.

    Helpful for things like implementing git flow etc.

    Attributes:
        name (str):
            The full name of the branch as returned by git.
        type (str):
            The type of the branch. This assumes the ``type/title`` branch
            naming format and will be the text before ``/`` in the branch name.
        title (str):
            The title of the branch. This assumes the ``type/title`` branch
            naming format and will be the text after ``/`` in the branch name.
    """
    name: str
    type: Optional[str] = None
    title: Optional[str] = None

    @classmethod
    def parse(cls, branch_name: str) -> 'BranchDetails':
        """ Parse branch name into BranchDetails instance.

        Args:
            branch_name (str):
                The name of the branch to parse.

        Returns:
            BranchDetails: The parsed branch name - easy to use.
        """
        if '/' in branch_name:
            branch_type, branch_title = branch_name.rsplit('/', 1)
            return BranchDetails(branch_name, branch_type, branch_title)
        else:
            return BranchDetails(branch_name, None, branch_name)


@dataclasses.dataclass
class CommitDetails:
    """ Allows querying git commits for details.

    Attributes:
        id (str):
            The commit ID (first 7 characters taken from `sha1`).
        sha1 (str):
            Full SHA1 of the commit.
        author (Author):
            The commit author.
        title (str):
            The title of the commit. This is the first line of the commit
            message and should usually be shorter than 50 characters.
        desc (str):
            The commit description. This is the portion following the title.
            Will be an empty string if the commit contains only the title.
        parents_sha1 (List[str]):
            List of SHA1s of parents of this commit.
    """
    sha1: str
    author: Author
    title: str
    desc: str
    parents_sha1: Optional[str]
    _branches: List[str] = dataclasses.field(default_factory=list)
    _parents: List['CommitDetails'] = dataclasses.field(default_factory=list)

    @property
    def id(self) -> str:
        return self.sha1[:7]

    @property
    def branches(self) -> List[str]:
        """ List of all branches this commit is a part of. """
        if self._branches is None:
            cmd = 'git branch --contains {}'.format(self.sha1)
            out = shell.run(
                cmd,
                capture=True,
                never_pretend=True
            ).stdout.strip()
            self._branches = [x.strip('* \t\n') for x in out.splitlines()]

        return self._branches

    @property
    def parents(self) -> List['CommitDetails']:
        """ Parents of the this commit. """
        if self._parents is None:
            self._parents = [CommitDetails.get(x) for x in self.parents_sha1]

        return self._parents

    # TODO: remove this property, only used by GAE plugin, which is dead anyway.
    @property
    def number(self) -> int:
        """ Return this commits number.

        This is the same as the total number of commits in history up until
        this commit.

        This value can be useful in some CI scenarios as it allows to track
        progress on any given branch (although there can be two commits with the
        same number existing on different branches).

        Returns:
            int: The commit number/index.
        """
        cmd = 'git log --oneline {}'.format(self.sha1)
        out = shell.run(cmd, capture=True, never_pretend=True).stdout.strip()
        return len(out.splitlines())

    @classmethod
    def get(cls, sha1: str = '') -> 'CommitDetails':
        """ Return details about a given commit.

        Args:
            sha1 (str):
                The sha1 of the commit to query. If not given, it will return
                the details for the latest commit.

        Returns:
            CommitDetails: Commit details. You can use the instance of the
            class to query git tree further.
        """
        with conf.within_proj_dir():
            cmd = 'git show -s --format="%H||%an||%ae||%s||%b||%P" {}'.format(
                sha1
            )
            result = shell.run(cmd, capture=True, never_pretend=True).stdout

        parts = result.split('||')
        if len(parts) != 6:
            raise ValueError('Invalid git show result:\n  {}'.format(result))

        sha1, name, email, title, desc, parents = result.split('||')

        return CommitDetails(
            sha1=sha1,
            author=Author(name, email),
            title=title,
            desc=desc,
            parents_sha1=parents.split(),
        )



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.versioning
    :synopsis: Functionality related to project versioning.
"""
import re
from typing import List, Optional, Tuple

from peltak.core import conf

from .version_file import (  # noqa: F401
    NodeVersionFile,
    PoetryVersionFile,
    PyVersionFile,
    RawVersionFile,
    VersionFile,
    is_valid,
    load_version_file,
)


# MAJOR.MINOR[.PATCH[-BUILD]]
RE_VERSION = re.compile(
    r'^'
    r'(?P<major>\d+)'
    r'(\.(?P<minor>\d+)'
    r'(\.(?P<patch>\d+))?)?'
    r'$'
)


def current() -> str:
    """ Return the project's current version.

    The project is read from the first file on the 'version.files' list. We
    always use the first file on the list as the source of truth, but any
    changes will be written to all files specified on the list.

    Returns:
        str: The current project version in MAJOR.MINOR.PATCH format. PATCH
        might be omitted if it's 0, so 1.0.0 becomes 1.0 and 0.1.0 becomes 0.1.
    """
    version_files = get_version_files()
    if not version_files:
        raise ValueError("No version file configured")

    main_version_file = version_files[0]
    return main_version_file.read() or ''


def write(version: str) -> None:
    """ Write the given version to the VERSION_FILE """
    if not is_valid(version):
        raise ValueError("Invalid version: '{}'".format(version))

    version_files = get_version_files()
    if not version_files:
        raise ValueError("No version file configured")

    for version_file in version_files:
        version_file.write(version)


def bump(component: str = 'patch', exact: Optional[str] = None) -> Tuple[str, str]:
    """ Bump the given version component.

    Args:
        component (str):
            What part of the version should be bumped. Can be one of:

            - major
            - minor
            - patch

        exact (str):
            The exact version that should be set instead of bumping the current
            one.

    Returns:
        tuple(str, str): A tuple of old and bumped version.
    """
    old_ver = current()

    if exact is None:
        new_ver = _bump_version(old_ver, component)
    else:
        new_ver = exact

    write(new_ver)
    return old_ver, new_ver


def _bump_version(version: str, component: str = 'patch') -> str:
    """ Bump the given version component.

    Args:
        version (str):
            The current version. The format is: MAJOR.MINOR[.PATCH].
        component (str):
            What part of the version should be bumped. Can be one of:

            - major
            - minor
            - patch

    Returns:
        str: Bumped version as a string.
    """
    if component not in ('major', 'minor', 'patch'):
        raise ValueError("Invalid version component: {}".format(component))

    m = RE_VERSION.match(version)
    if m is None:
        raise ValueError(
            f"Version must be in MAJOR.MINOR[.PATCH] format, got: '{version}'"
        )

    major = m.group('major')
    minor = m.group('minor') or '0'
    patch = m.group('patch') or None

    if component == 'major':
        major = str(int(major) + 1)
        minor = '0'
        patch = '0'

    elif component == 'minor':
        minor = str(int(minor) + 1)
        patch = '0'

    else:
        patch = patch or '0'
        patch = str(int(patch) + 1)

    new_ver = '{}.{}.{}'.format(major, minor, patch)

    return new_ver


def get_version_files() -> List[VersionFile]:
    version_files = conf.get('version.files', [])

    if not version_files:
        # TODO: 'version.file' is deprecated, use 'version.files' instead.
        single = conf.get('version.file', None)
        version_files = [single] if single else []

    return [load_version_file(conf.proj_path(p)) for p in version_files]



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Application entry point. """

# Scripts should be available by default
import peltak.commands.scripts  # noqa: F401 pylint: disable=unused-import
# Make sure config is loaded
from peltak.commands import root_cli
from peltak.core import context  # noqa: F401 pylint: disable=unused-import
from peltak.core import conf


__all__ = [
    'root_cli'
]


# Accessing the config for the first time will load it.
# This is crucial for the completion to work well. We need to load the config
# here so we have autocompletion for all commands defined in the config.
conf.init()


from peltak.commands.root import clean  # noqa: F401, E402 pylint: disable=unused-import



from unittest.mock import patch

import pytest

from peltak.core import conf


@pytest.fixture
def app_conf():
    appconf = conf.Config()
    with patch('peltak.core.conf.g_conf', appconf):
        yield appconf



import os
import os.path
import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Any, Dict, Iterator, Optional, Union

from . import hooks, log, util


g_conf: Optional['Config'] = None
ConfigDict = Dict[str, Any]
DEFAULT_PELCONF_NAME = 'pelconf.yaml'
PELTAK_CONFIG_FILES = frozenset({'pelconf.yaml', 'pyproject.toml'})


class ConfigError(RuntimeError):
    pass


class ConfigNotInitialized(ConfigError):
    pass


class ConfigAlreadyInitialized(ConfigError):
    pass


def init():
    global g_conf

    if g_conf:
        raise ConfigAlreadyInitialized()

    config_path = _discover_proj_config()
    g_conf = _load_config(config_path)

    hooks.register.call('post-conf-load')


def get(name: str, *default: Any) -> Any:
    global g_conf

    if not g_conf:
        raise ConfigNotInitialized()

    return g_conf.get(name, *default)


def get_path(name: str, *default: Any) -> Any:
    global g_conf

    if not g_conf:
        raise ConfigNotInitialized()

    return g_conf.get_path(name, *default)


def get_env(name: str, *default: Any) -> Union[str, Any]:
    global g_conf

    if not g_conf:
        raise ConfigNotInitialized()

    return g_conf.get_env(name, *default)


def proj_path(*path_parts: str) -> str:
    global g_conf

    if not g_conf:
        raise ConfigNotInitialized()

    return g_conf.proj_path(*path_parts)


def as_dict() -> ConfigDict:
    global g_conf

    if not g_conf:
        raise ConfigNotInitialized()

    return g_conf.values


@contextmanager
def within_proj_dir(path: str = '.') -> Iterator[None]:
    global g_conf

    if not g_conf:
        raise ConfigNotInitialized()

    curr_dir = os.getcwd()

    os.chdir(g_conf.proj_path(path))
    yield
    os.chdir(curr_dir)


class Config:
    """ Represents the `pelconf.yaml` file. """
    def __init__(
        self,
        values: Optional[ConfigDict] = None,
        *,
        path: Optional[str] = None,
    ):
        self.values: ConfigDict = values or {}
        self.path = path
        self.root_dir = path and os.path.dirname(path)

    def reset(self, config):
        """ Reset config to the given values.

        This will completely overwrite the current configuration.
        """
        self.values = config

    def has(self, name: str) -> bool:
        return util.dict_has(self.values, name)

    def get(self, name: str, *default: Any) -> Any:
        """ Get config value with the given name and optional default.

        Args:
            name (str):
                The name of the config value.
            *default (Any):
                If given and the key doesn't not exist, this will be returned
                instead. If it's not given and the config value does not exist,
                AttributeError will be raised

        Returns:
            The requested config value. This is one of the global values defined
            in this file. If the value does not exist it will return *default* if
            give or raise `AttributeError`.

        Raises:
            AttributeError: If the value does not exist and *default* was not given.
        """
        try:
            return util.get_from_dict(self.values, name, *default)
        except KeyError:
            raise AttributeError(f"Config value '{name}' does not exist")

    def get_path(self, name: str, *default: Any) -> Any:
        """ Get config value as path relative to the project directory.

        This allows easily defining the project configuration within the fabfile
        as always relative to that fabfile.

        Args:
            name (str):
                The name of the config value containing the path.
            *default (Any):
                If given and the key doesn't not exist, this will be returned
                instead. If it's not given and the config value does not exist,
                AttributeError will be raised

        Returns:
            The requested config value. This is one of the global values defined
            in this file. If the value does not exist it will return *default* if
            give or raise `AttributeError`.

        Raises:
            AttributeError: If the value does not exist and *default* was not given.
        """
        value = self.get(name, *default)

        if value is None:
            return None

        return self.proj_path(value)

    def get_env(self, name: str, *default: Any) -> Union[str, Any]:
        """ Get the value of an ENV variable. """
        return os.environ.get(name, *default)

    def proj_path(self, *path_parts: str) -> str:
        """ Return absolute path to the repo dir (root project directory).

        Args:
            *path_parts (str):
                The path relative to the project root (pelconf.yaml).

        Returns:
            str: The given path converted to an absolute path.
        """
        parts = list(path_parts) or ['.']

        # If path represented by path_parts is absolute, do not modify it.
        if not os.path.isabs(parts[0]):

            if self.root_dir is not None:
                parts = [self.root_dir] + list(parts)

        return os.path.normpath(os.path.join(*parts))

    @contextmanager
    def within_proj_dir(self, path: str = '.') -> Iterator[None]:
        """ Return an absolute path to the given project relative path.

        :param path:
            Project relative path that will be converted to the system wide absolute
            path.
        :return:
            Absolute path.
        """
        curr_dir = os.getcwd()

        os.chdir(self.proj_path(path))
        yield
        os.chdir(curr_dir)


def _load_config(path: str) -> Config:
    values = _load_from_file(path) if path else {}
    cfg = Config(values=values, path=path)

    # Prepend python_paths to sys.path. Using a clever slice notation to prepend
    # in one go.
    python_paths = [cfg.proj_path(p) for p in cfg.get('python_paths', [])]
    sys.path[0:0] = [p for p in python_paths if p not in sys.path]

    # Add scripts_dir to python paths so we can directly import all the python
    # scripts that exist there.
    scripts_dir = cfg.get_path('scripts_dir')
    if scripts_dir and scripts_dir not in sys.path:
        sys.path.insert(0, scripts_dir)

    # TODO: src_dir is deprecated, use 'python_paths' config value instead.
    if cfg.has('src_dir'):
        sys.path.insert(0, cfg.get_path('src_dir'))

    for cmd in cfg.get('commands', []):
        try:
            py_import(cmd)
        except ImportError as ex:
            log.err("Failed to load commands from <33>{}<31>: {}", cmd, ex)

    return cfg


def _load_from_file(proj_config: str):
    """ Load configuration from file. """
    if proj_config.endswith('pyproject.toml'):
        return _load_from_toml_file(proj_config)
    elif proj_config.endswith('pelconf.yaml'):
        return _load_from_yaml_file(proj_config)
    else:
        raise RuntimeError(f"Unsupported configuration {proj_config}")


def _load_from_yaml_file(path):
    """ Load config values from a YAML file. """
    with open(path) as fp:
        values = util.yaml_load(fp) or {}
        return values


def _load_from_toml_file(path):
    """ Load config values from a YAML file. """
    config = util.toml_load(path)
    values = config.get('tool', {}).get('peltak', {})
    return values


# @util.cached_result()
def _discover_proj_config() -> Optional[str]:
    """ Find the project path by going up the file tree.

    This will look in the current directory and upwards for pelconf.yaml  or
    pyproject.toml.
    """
    curr = os.getcwd()

    while curr.startswith('/') and len(curr) > 1:
        files = frozenset(os.listdir(curr))
        config_files = files & PELTAK_CONFIG_FILES
        if config_files:
            if DEFAULT_PELCONF_NAME in config_files:
                # If pelconf.yaml is present, use it. This handles both cases
                # where the project has both pelconf.yaml and pyproject.toml or
                # when it only has pelconf.yaml.
                return os.path.join(curr, DEFAULT_PELCONF_NAME)
            else:
                # If we don't have pelconf.yaml, it means we only have
                # pyproject.toml so we will use it.
                return os.path.join(curr, next(iter(config_files)))
        else:
            curr = os.path.dirname(curr)

    return None


def py_import(cmd: str) -> ModuleType:
    """ Exists only so we can patch it in tests."""
    return __import__(cmd)   # nocov



# Copyright 2017-2021 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.git
    :synopsis: Git helpers.
"""
from .branch import (  # noqa:  F401
    branches,
    commit_branches,
    commit_details,
    current_branch,
    guess_base_branch,
    latest_commit,
    protected_branches,
    verify_branch,
)
from .status import staged, unstaged, untracked  # noqa:  F401
from .types import Author, BranchDetails, CommitDetails  # noqa:  F401
from .util import commit_author, config, ignore, tag, tags  # noqa:  F401



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Engine wraps the jinja2 environment and exposes it to the rest of the code. """
from typing import Any, Dict

import jinja2

from peltak.core import util

from . import filters


class Engine(util.Singleton):
    """ Template engine wrapper.

    Engine is a singleton, which means there is always a single instance
    in existence during the app run. You can access this instance by calling
    the class initializer:

    >>> from peltak.core.templates import Engine
    >>>
    >>> engine = Engine()


    Only on first call will the class be actually constructed, all following
    calls will return the same instance:

    >>> from peltak.core.templates import Engine
    >>>
    >>> Engine() is Engine()
    True

    """
    def __init__(self):
        if not self._singleton_initialized:
            self.env = self._make_env()
            pass

    def render(self, template_str: str, template_ctx: Dict[str, Any] = None) -> str:
        """ Render a script template using the given context.

        Examples:

            >>> from peltak.core.templates import Engine
            >>>
            >>> Engine().render("{{ msg | upper }}", {'msg': 'hello'})
            'HELLO'

        """
        return self.env.from_string(template_str).render(template_ctx)

    def render_file(self, template_file: str, template_ctx: Dict[str, Any] = None):
        """ Render a template file from src/peltak/templates directory.

        All built-in peltak templates should go to the ``src/peltak/templates``
        directory.

        Args:
            template_file (str):
                The name of the template file to use. This file needs to reside
                in the templates directory and will be automatically loaded.
            template_ctx (dict[str, Any]):
                The template context to use. This is a dictionary of values that
                will be injected into the template during rendering.

        Returns:
            str: A rendered template.

        Examples:

            \b
            >>> from peltak.core.templates import Engine
            >>>
            >>> output = Engine().render_file("pelconf.yaml", {'src_dir': 'src'})
            >>> print(output)
            # peltak configuration file
            # Visit https://novopl.github.io/peltak for more information
            pelconf_version: '1'
            <BLANKLINE>
            # You can add custom project commands or 3rd party packages here.
            commands:
              - peltak.extra.scripts
              - peltak.extra.git
              - peltak.extra.gitflow
            <BLANKLINE>
            <BLANKLINE>
            <BLANKLINE>
            # This directory will be added to sys.path when the config is loaded.
            # Useful if do not keep the source code in th root directory.
            python_paths: ['src']
            <BLANKLINE>
            <BLANKLINE>
            <BLANKLINE>
            # Scripts to help you manage your project.
            scripts:
              test:
                about: Test your code
                command: |
                  echo "Change me, I'm just a test command"

        """
        template = self.env.get_template(template_file)
        return template.render(template_ctx)

    def _make_env(self) -> jinja2.Environment:
        """ Initialize jinja2 env. """
        env = jinja2.Environment(
            loader=jinja2.PackageLoader('peltak', 'templates'),
            variable_start_string='{{',
            variable_end_string='}}',
            trim_blocks=True,
            lstrip_blocks=True,
        )

        env.filters['header'] = filters.header_filter
        env.filters['count_flag'] = filters.count_flag_filter
        env.filters['cprint'] = filters.cprint_filter
        env.filters['wrap_paths'] = filters.wrap_paths_filter

        return env



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
#####################
Create peltak scripts
#####################

.. module: peltak.extra.scripts
    :synopsis: Custom project scripts

**peltak** supports defining simple scripts directly inside
:ref:`pelconf.yaml <peltak.core.conf>`. On top of just defining the command, by
default the command is processed by jinja before it's being ran. This makes it
possible to inject some dynamic values into the command and makes the whole
scripts subsystem very flexible.


Get a list of scripts available in the project
==============================================

To get a list of all scripts that are defined in :ref:`pelconf.yaml <peltak.core.conf>`
you only need to run the ``peltak run`` command without any arguments.
Here's an example from the **peltak** project itself (at the time ofwriting)::

    $ peltak run
    Usage: peltak run [OPTIONS] COMMAND [ARGS]...

      Run custom scripts

    Options:
      --help  Show this message and exit.

    Commands:
      checks       Run all checks (types, pep8, code style)
      cov-core     Generate coverage report for peltak.core
      cov-extra    Generate coverage report for peltak.extra
      cov-scripts  Generate coverage report for peltak.extra.scripts
      docs         Generate sphinx documentation
      tests        Run unit tests
      type-check   Run static type checks


Defining a script
=================

You can define a script inside :ref:`pelconf.yaml <peltak.core.conf>` in the
**scripts:**. The script always has a name (key in the scripts config section)
and a command (required prop). Say you want to define a ``test`` command that
invokes pytest on the ``./test`` directory, here's the corresponding
:ref:`pelconf.yaml <peltak.core.conf>` section.

.. code-block:: yaml

    scripts:
        test:
            command: pytest ./test

And you can run this command with:

.. code-block:: bash

    peltak run test


Debugging your script
~~~~~~~~~~~~~~~~~~~~~

**peltak** provides a few helpers to ease the process of creating custom
project scripts. The first one is the ``--pretend`` option. Providing it when
invoking your script will cause **peltak** to render the compiled template
script instead of running it. This allows you to see exactly what command would
be executed.

The second one is the ``-v``, ``--verbose`` flag. With verbosity level 3 (``-vvv``)
**peltak** will render the full script template context as a highlighted YAML.
This makes it very easy to see exactly what values are available in the template
context of your project.

Combining ``--pretend`` with ``-vvv`` will show you the full template context
followed by a compiled command without executing anything. This is perfect when
you're working on your script and are afraid you might break something.


Multiline and multi statement commands
======================================

The *test* command above is a very simple one. What if your script command is
very long, or you want to call multiple shell commands? The command you define
is treated as a shell script, so you can do whatever you can with shell scripts.
One thing to remember is to use ``set -e`` in your script if you want it to fail
if any of the commands fail.

For convenience, you can use the YAML multiline string syntax:

.. code-block:: yaml

    checks:
        about: Run checks on the code base
        command: |
            set -e
            pipenv run mypy src
            pipenv run pep8 --config tools/pep8.ini src
            pipenv run pylint --rcfile tools/pylint.ini src


Templating capabilities of scripts
==================================

The scripts module was designed to parse the commands as templates and inject
**peltak's** current working environment. This means you can use configuration
and some helper filters to generate the actual command.

If you want to more details about templating in scripts, you can read
:ref:`peltak.core.templates`.

.. note::
    Scripts templating is very simple. It should be used only for simple
    value substitution with optional processing. If need flow control for a
    script you probably should look at implementing a custom peltak command for
    the project as this will give you the full power of python and is also a
    very simple to do.


Injecting config values from :ref:`pelconf.yaml <peltak.core.conf>` into your script
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Let's modify our *checks* command to use config variables. We're repeating
ourselves passing ``src`` to every command. You probably already have a
``src_dir`` variable set up in your config, why not use that to execute the
checks? Same goes for the location of the tools config files. We can create
a new config variable to store the tool config dir and just use it in
our command. The current configuration is injected into the template as ``conf``
object. You can access it via the ``{{ EXPRESSION }}`` syntax. Here's a
modified version of the *checks* command.

.. code-block:: python

    src_dir: src
    tool_dir: tools

    checks:
      about: Run checks on the code base
      command: |
        pipenv run mypy {{ conf.src_dir }}
        pipenv run pep8 --config {{ conf.tool_dir }}/pep8.ini {{ conf.src_dir }};
        pipenv run pylint --rcfile {{ conf.tool_dir }}/pylint.ini {{ conf.src_dir }};

This makes it easy to reuse certain values between scripts as all scripts can
access config through ``{{ conf.VALUE }}`` expression. This makes managing the
project much easier as :ref:`pelconf.yaml <peltak.core.conf>` can serve as a
config store for all scripts and changing any of those values will work right
away with all scripts that use it.

You can find out more in `/reference/script_filters`

"""



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" scripts logic. """
import dataclasses
import subprocess
import sys

import yaml

from peltak.core import conf, fs, log, shell, templates
from peltak.core.context import RunContext

from .types import CliOptions, Script


def run_script(script: Script, options: CliOptions) -> None:
    """ Run the script with the given (command line) options. """
    template_ctx = build_template_context(script, options)
    verbose = RunContext().get('verbose')
    pretend = RunContext().get('pretend')

    if verbose >= 3:
        log.info('Compiling script <35>{name}\n{script}'.format(
            name=script.name,
            script=shell.highlight(script.command, 'jinja')
        ))
        yaml_str = yaml.dump(template_ctx, default_flow_style=False)
        log.info('with context:\n{}\n'.format(shell.highlight(yaml_str, 'yaml')))

    # Command is either specified directly in pelconf.yaml or lives in a
    # separate file.
    command = script.command
    if script.command_file:
        with open(conf.proj_path(script.command_file)) as fp:
            command = fp.read()

    if not command:
        raise ValueError("Scripts must have 'command' or 'command_file' specified.")

    cmd = templates.Engine().render(command, template_ctx)
    retcode = exec_script_command(cmd, pretend)

    if verbose:
        log.info("Script exited with code: <33>{}", retcode)

    if retcode not in script.success_exit_codes:
        sys.exit(retcode)


def exec_script_command(cmd: str, pretend: bool) -> int:
    """ This will execute the already compiled script command.

    This function serves the purpose of encapsulating the low level code of
    spawning a subprocess from the rest of the logic.
    """
    if not pretend:
        with conf.within_proj_dir():
            # God knows why, if we run the command using `shell.run()` and it
            # exists with non-zero code it will also kill the parent peltak
            # process. The Popen arguments are the same, to my knowledge
            # everything is the same but the way it behaves is completely
            # different. If we just use Popen directly, everything works as
            # expected  ¯\_(ツ)_/¯
            # TODO: This possibly happens because of exit_on_error in shell.run()
            p = subprocess.Popen(cmd, shell=True)
            try:
                p.communicate()
                return p.returncode
            except KeyboardInterrupt:
                p.kill()
                return -1
    else:
        log.info(
            "<90>{bar}<0>\n{script}\n<90>{bar}",
            bar='=' * 80,
            script=shell.highlight(cmd, 'bash'),
        )
        return 0


def build_template_context(script, options):
    """ Build command template context.

    This will collect all the values like current configuration, command line
    options, runtime context etc. and pass it to the script template.
    """
    template_ctx = {
        'opts': dict(
            verbose=RunContext().get('verbose'),
            pretend=RunContext().get('pretend'),
            **options
        ),
        'script': dataclasses.asdict(script),
        'conf': conf.as_dict(),
        'ctx': RunContext().values,
        'proj_path': conf.proj_path,
    }

    if script.files:
        template_ctx['files'] = fs.collect_files(script.files)

    return template_ctx



""" A place to store all patch functions specific to peltak.

Easier to track them down, when you need one.
"""
from functools import wraps
from typing import Any, Optional
from unittest.mock import Mock, mock_open, patch

from peltak.core import conf, shell
from peltak.core.shell import ExecResult


def patch_is_tty(value):
    """ Wrapped test function will have peltak.core.shell.is_tty set to *value*. """
    def decorator(fn):  # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):   # pylint: disable=missing-docstring
            is_tty = shell.is_tty
            shell.is_tty = value

            try:
                return fn(*args, **kw)
            finally:
                shell.is_tty = is_tty
        return wrapper

    return decorator


def patch_proj_root(path):
    """ Overwrite the project root path in conf.Config singleton. """
    from peltak.core import conf

    def decorator(fn):  # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):  # pylint: disable=missing-docstring
            curr_root = conf.g_conf.root_dir
            conf.root_dir = path

            result = fn(*args, **kw)

            conf.root_dir = curr_root
            return result

        return wrapper
    return decorator


def patch_pelconf(
    values: Optional[conf.ConfigDict] = None,
    *,
    path: str = '/fake/proj/pelconf.yaml',
):
    return patch('peltak.core.conf.g_conf', conf.Config(
        values=values,
        path=path,
    ))


def patch_run(
    stdout: Optional[str] = None,
    retcode: Optional[int] = None,
    stderr: Optional[str] = None,
    cmd: Optional[str] = None
) -> Any:
    """ Patch shell.run and make it return a given result.

    Args:
        stdout (str):
        retcode (int):
        stderr (str):
        cmd (str):

    Returns:
        FunctionType:

    """
    p_run = Mock(return_value=ExecResult(
        cmd or '',
        retcode or 0,
        stdout or '',
        stderr or '',
        retcode == 0,
        retcode != 0,
    ))

    return patch('peltak.core.shell.run', p_run)


def patch_open(module: str, read_data: str = '', create: bool = True) -> Any:
    """ Patch builtin open() function for the given module

    This is a convenient wrapper around mock_open.
    """
    return patch(
        module + '.open',
        create=create,
        new_callable=mock_open,
        read_data=read_data,
    )



# Copyright 2017-2021 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from typing import Any, Dict, List, Optional

from .. import conf, shell, util
from .branch import latest_commit
from .types import Author


@util.mark_deprecated('CommitDetails.get().author')
def commit_author(sha1: str = '') -> Author:
    """ Return the author of the given commit.

    Args:
        sha1 (str):
            The sha1 of the commit to query. If not given, it will return the
            sha1 for the current commit.
    Returns:
        Author: A named tuple ``(name, email)`` with the commit author details.
    """
    with conf.within_proj_dir():
        cmd = 'git show -s --format="%an||%ae" {}'.format(sha1)
        result = shell.run(
            cmd,
            capture=True,
            never_pretend=True
        ).stdout
        name, email = result.split('||')
        return Author(name, email)


@util.cached_result()
def ignore() -> List[str]:
    """ Return a list of patterns in the project .gitignore

    Returns:
        list[str]: List of patterns set to be ignored by git.
    """

    def parse_line(line):   # pylint: disable=missing-docstring
        # Decode if necessary
        if not isinstance(line, str):
            line = line.decode('utf-8')

        # Strip comment
        line = line.split('#', 1)[0].strip()

        return line

    ignore_files = [
        conf.proj_path('.gitignore'),
        conf.proj_path('.git/info/exclude'),
        config().get('core.excludesfile')
    ]

    result: List[str] = []
    for ignore_file in ignore_files:
        if not (ignore_file and os.path.exists(ignore_file)):
            continue

        with open(ignore_file) as fp:
            parsed = (parse_line(line) for line in fp.readlines())
            result += [x for x in parsed if x]

    return result


def tag(name: str, message: str, author: Optional[Author] = None):
    """ Tag the current commit.

    Args:
        name (str):
            The tag name.
        message (str):
            The tag message. Same as ``-m`` parameter in ``git tag``.
        author (Author):
            The commit author. Will default to the author of the commit.
        pretend (bool):
            If set to **True** it will print the full ``git tag`` command
            instead of actually executing it.
    """
    cmd = (
        'git -c "user.name={author.name}" -c "user.email={author.email}" '
        'tag -a "{name}" -m "{message}"'
    ).format(
        author=author or latest_commit().author,
        name=name,
        message=message.replace('"', '\\"').replace('`', '\\`'),
    )
    shell.run(cmd)


@util.cached_result()
def config() -> Dict[str, Any]:
    """ Return the current git configuration.

    Returns:
        dict[str, Any]: The current git config taken from ``git config --list``.
    """
    out = shell.run(
        'git config --list',
        capture=True,
        never_pretend=True
    ).stdout.strip()

    result = {}
    for line in out.splitlines():
        name, value = line.split('=', 1)
        result[name.strip()] = value.strip()

    return result


@util.cached_result()
def tags() -> List[str]:
    """ Returns all tags in the repo.

    Returns:
        list[str]: List of all tags in the repo, sorted as versions.

    All tags returned by this function will be parsed as if the contained
    versions (using ``v:refname`` sorting).
    """
    return shell.run(
        'git tag --sort=v:refname',
        capture=True,
        never_pretend=True
    ).stdout.strip().splitlines()



import dataclasses
from typing import Any, Dict, Optional, Tuple


@dataclasses.dataclass
class ExcValues:
    args: Tuple
    kwargs: Dict[str, Any]


class PeltakError(Exception):
    msg: str = "peltak error"

    def __init__(
        self,
        detail: Optional[str] = None,
        *args,
        **kw
    ):
        message = self.msg.format(*args, **kw) + (f": {detail}" if detail else '')
        super(PeltakError, self).__init__(message)

        self.detail = detail
        # self.args is already used by Exception
        self.values = ExcValues(args, kw)



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
##############
``peltak git``
##############

git helpers.
"""
from peltak.commands import click, pretend_option, root_cli


@root_cli.group('git')
def git_cli():
    """ Git related commands """
    pass


@git_cli.command('add-hooks')
@pretend_option
@click.option(
    '--pre-commit',
    type=str,
    prompt='Pre commit tool',
    help=(
        'Command that will be ran before every commit. If it fails, the commit '
        'will fail.'
    )
)
@click.option(
    '--pre-push',
    type=str,
    prompt='Pre push tool',
    help=(
        'Command that will be ran before every push. If it fails, the push '
        'will fail.'
    )
)
def add_hooks(pre_commit: str, pre_push: str):
    """ Setup project git hooks.

    This will run all the checks before pushing to avoid waiting for the CI
    fail.

    Example::

        $ peltak git add-hooks

    """
    from . import git_impl

    git_impl.add_hooks(pre_commit, pre_push)


@git_cli.command('push')
@pretend_option
def push():
    """ Push the current branch and set to track remote.

    Example::

        $ peltak git push --pretend
        $ peltak git push

    """
    from . import git_impl
    git_impl.push()


@git_cli.command('delete-remote')
@pretend_option
def delete_remote():
    """ Delete the current branch on remote origin.

    This is an equivalent of ``git push origin :<branch>``. Easy way to quickly
    delete the remote branch without having to type in the branch name.

    Example::

        $ peltak git delete-remote --pretend
        $ peltak git delete-remote

    """
    from . import git_impl
    git_impl.delete_remote()



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.util
    :synopsis: Various helpers that do not depend on anything else in the project.
"""
import re
import time
import warnings
from functools import wraps
from pprint import pformat
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Text,
    TextIO,
    Union,
)

import tomlkit
import yaml


try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader  # type: ignore


TextOrStream = Union[Text, TextIO]
PlainDict = Dict[str, Any]
YamlData = Union[PlainDict, List[Any]]
AnyFunction = Callable[..., Any]
Decorator = Callable[[AnyFunction], AnyFunction]


class timed_block(object):  # noqa
    """ Context manager to measure execution time for a give block of code.

    Example:
        >>> import time
        >>> from peltak.core import util
        >>>
        >>> with util.timed_block() as t:
        ...     time.sleep(1)
        >>>
        >>> print("Code executed in {}s".format(int(t.elapsed_s)))
        Code executed in 1s

    Attributes:
        t0 (float):
            The time at the start of execution.
        elapsed (float);
            Raw elapsed time in seconds.
        elapsed_s (float):
            Elapsed time in seconds, rounded to 3 decimal places for easy
            display.
        elapsed_ms (float):
            Elapsed time in milliseconds, rounded to 3 decimal places for easy
            display.
    """
    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed = time.time() - self.t0
        self.elapsed_s = round(self.elapsed, 3)
        self.elapsed_ms = round(self.elapsed * 1000, 3)


def mark_experimental(fn: AnyFunction) -> AnyFunction:
    """ Mark function as experimental.

    Args:
        fn (FunctionType):
            The command function to decorate.
    """
    @wraps(fn)
    def wrapper(*args, **kw):   # pylint: disable=missing-docstring
        from peltak.core import shell

        if shell.is_tty:
            warnings.warn("This command is has experimental status. The "
                          "interface is not yet stable and might change "
                          "without notice. Use at your own risk")
        return fn(*args, **kw)

    return wrapper


def mark_deprecated(replaced_by: str) -> Decorator:
    """ Mark command as deprecated.

    Args:
        replaced_by (str):
            The command that deprecated this command and should be used instead.
    """
    def decorator(fn):   # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):   # pylint: disable=missing-docstring
            from peltak.core import shell

            if shell.is_tty:
                warnings.warn("This command is has been deprecated. Please use "
                              "'{new}' instead.".format(new=replaced_by))

            return fn(*args, **kw)

        return wrapper

    return decorator


class cached_result(object):
    """ Decorator that caches the function result.

    This is especially useful for functions whose output won't change during
    a single *peltak* execution run and are expensive (i.e. call external
    shell commands).

    Example:
        >>> from peltak.core import util
        >>>
        >>> @util.cached_result()
        ... def foo():
        ...     call_count = getattr(foo, 'call_count', 0)
        ...     call_count += 1
        ...     setattr(foo, 'call_count', call_count)
        >>> foo()
        >>> print(foo.call_count)
        1
        >>> foo()
        >>> print(foo.call_count)
        1
        >>> util.cached_result.clear(foo)
        >>> foo()
        >>> print(foo.call_count)
        2
    """
    CACHE_VAR = '__cached_result'

    def __init__(self):
        pass

    def __call__(self, fn: AnyFunction) -> AnyFunction:
        """ Apply the decorator to the given function.

        Args:
            fn (FunctionType):
                The function to decorate.
        :return Function:
            The function wrapped in caching logic.
        """
        @wraps(fn)
        def wrapper(refresh=False):   # pylint: disable=missing-docstring
            if refresh or not hasattr(wrapper, self.CACHE_VAR):
                result = fn()
                setattr(wrapper, self.CACHE_VAR, result)

            return getattr(wrapper, self.CACHE_VAR)

        return wrapper

    @classmethod
    def clear(cls, fn: AnyFunction):
        """ Clear result cache on the given function.

        If the function has no cached result, this call will do nothing.

        Args:
            fn (FunctionType):
                The function whose cache should be cleared.
        """
        if hasattr(fn, cls.CACHE_VAR):
            delattr(fn, cls.CACHE_VAR)


def in_batches(iterable: Iterable[Any], batch_size: int) -> Iterator[List[Any]]:
    """ Split the given iterable into batches.

    Args:
        iterable (Iterable[Any]):
            The iterable you want to split into batches.
        batch_size (int):
            The size of each bach. The last batch will be probably smaller (if
            the number of elements cannot be equally divided.

    Returns:
        Iterator[list[Any]]: Will yield all items in batches of **batch_size**
            size.

    Example:

        >>> from peltak.core import util
        >>>
        >>> batches = util.in_batches([1, 2, 3, 4, 5, 6, 7], 3)
        >>> batches = list(batches)     # so we can query for length
        >>> len(batches)
        3
        >>> batches
        [[1, 2, 3], [4, 5, 6], [7]]

    """
    items = list(iterable)
    size = len(items)

    for i in range(0, size, batch_size):
        yield items[i:min(i + batch_size, size)]


def yaml_load(str_or_fp: TextOrStream) -> YamlData:
    """ Load data from YAML string or file-like object.

    Args:
        str_or_fp (Union[str, TextIO]):

    Returns:
        YamlData: The data loaded from the YAML string/file.
    """
    return yaml.load(str_or_fp, Loader=Loader)


def yaml_dump(data: YamlData, stream: Optional[Any] = None) -> Text:
    """ Dump data to a YAML string/file.

    Args:
        data (YamlData):
            The data to serialize as YAML.
        stream (TextIO):
            The file-like object to save to. If given, this function will write
            the resulting YAML to that stream.

    Returns:
        str: The YAML string.
    """
    return yaml.dump(
        data,
        stream=stream,
        Dumper=Dumper,
        default_flow_style=False
    )


def toml_load(path_or_fp: TextOrStream) -> PlainDict:
    """ Load TOML configuration into a dict. """
    if isinstance(path_or_fp, str):
        with open(path_or_fp) as fp:
            document = tomlkit.parse(fp.read())
            return dict(document)
    else:
        document = tomlkit.parse(path_or_fp.read())
        return dict(document)


def toml_dump(data: PlainDict, path_or_fp: Optional[TextOrStream] = None):
    """ Save a plain dict as a TOML file. """
    document = tomlkit.item(data)
    if path_or_fp is None:
        return tomlkit.dumps(document)
    elif isinstance(path_or_fp, str):
        with open(path_or_fp, 'w') as fp:
            fp.write(tomlkit.dumps(document))

    else:
        path_or_fp.write(tomlkit.dumps(document))


def remove_indent(text: str) -> str:
    """ Remove indentation from the text.

    All indentation will be removed no matter if it's consistent across the
    *text*.

    Args:
        text (str): The text that contains indentation.

    Returns:
        str: The same text but with all indentation removed from each line.
    """
    return re.sub('\n[ \t]+', '\n', text.lstrip())


class Singleton(object):
    """ Base class for singletons.

    Inheriting from this class will make your class a singleton. It should have
    an initializer that takes no arguments.
    """
    instances: Dict[str, Any] = {}

    def __new__(cls, *args, **kw):
        # Could upgrade this to thread local storage
        instance = Singleton.instances.get(cls.__name__)

        if instance:
            instance._singleton_initialized = True
        else:
            instance = object.__new__(cls)
            instance._singleton_initialized = False
            Singleton.instances[cls.__name__] = instance

        return instance


def get_from_dict(dct: Dict, path: str, *default: Any) -> Any:
    """ Get value from dictionary using a dotted path.

    This allows you to get a value from deep within a dictionary without having
    to manually check at each level if the key exists and handling the default
    properly. This function will work the same as dict.get() but allows you to
    specify a path in dotted notation, eg: 'main.sub.value'. Depending on whether
    you passed a default value it will either return it or raise KeyError
    if any of the keys or value itself is missing.

    Args:
        name (str):
            The name of the config value.
        *default (Any):
            If given and the key doesn't not exist, this will be returned
            instead. If it's not given and the config value does not exist,
            KeyError will be raised.

    Returns:
        The requested config value. This is one of the global values defined
        in this file. If the value does not exist it will return *default* if
        give or raise `AttributeError`.

    Raises:
        KeyError: If the value does not exist and *default* was not given.

    Examples:
        >>> d = {
        ...     'main': {
        ...         'value': 123,
        ...     }
        ... }
        ...
        >>> get_from_dict(d, 'main.value')
        123
        >>> get_from_dict(d, 'missing', 321)
        321
        >>> get_from_dict(d, 'missing')
        Traceback (most recent call last):
        KeyError:
    """
    curr = dct
    for part in path.split('.'):
        if part in curr:
            curr = curr[part]
        elif default:
            return default[0]
        else:
            raise KeyError(f"'{path}' not found in dict: {pformat(dct)}")

    return curr


def set_in_dict(dct: Dict, path: str, value: Any) -> None:
    """ Set value in a dictionary using a dotted path.

    This is the opposite to `get_from_dict()` function, but sets the value under
    the given path instead.


    Args:
        name (str):
            The name of the context value to change.
        value (Any):
            The new value for the selected context value

    Raises:
        KeyError: If the value does not exist and *default* was not given.

    Examples:
        >>> d = {
        ...     'main': {
        ...         'value': 123,
        ...     }
        ... }
        ...
        >>> set_in_dict(d, 'main.value', 321)
        >>> d
        {'main': {'value': 321}}

    """
    curr = dct
    parts = path.split('.')

    for i, part in enumerate(parts[:-1]):
        try:
            curr = curr.setdefault(part, {})
        except AttributeError:
            raise KeyError('.'.join(parts[:i + 1]))

    try:
        curr[parts[-1]] = value
    except TypeError:
        raise KeyError('.'.join(parts[:-1]))


def dict_has(dct: Dict, path: str) -> bool:
    curr = dct
    for part in path.split('.'):
        if part in curr:
            curr = curr[part]
        else:
            return False

    return True



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Git commands implementation. """
import os

import click

from peltak.core import conf, context, fs, git, log, shell, util


def add_hooks(pre_commit: str, pre_push: str):
    """ Add git hooks for commit and push to run linting and tests. """

    # Detect virtualenv the hooks should use

    # Detect virtualenv
    virtual_env = conf.get_env('VIRTUAL_ENV')
    if virtual_env is None:
        log.err("You are not inside a virtualenv")
        confirm_msg = (
            "Are you sure you want to use global python installation "
            "to run your git hooks? [y/N] "
        )
        click.prompt(confirm_msg, default='')
        if not click.confirm(confirm_msg):
            log.info("Cancelling")
            return

        load_venv = ''
    else:
        load_venv = 'source "{}/bin/activate"'.format(virtual_env)

    commit_hook = conf.proj_path('.git/hooks/pre-commit')
    push_hook = conf.proj_path('.git/hooks/pre-push')

    # Write pre-commit hook
    log.info("Adding pre-commit hook <33>{}", commit_hook)
    fs.write_file(commit_hook, util.remove_indent('''
        #!/bin/bash
        PATH="/opt/local/libexec/gnubin:$PATH"
        
        {load_venv}
        
        {command}
        
    '''.format(load_venv=load_venv, command=pre_commit)))

    # Write pre-push hook
    log.info("Adding pre-push hook: <33>{}", push_hook)
    fs.write_file(push_hook, util.remove_indent('''
        #!/bin/bash
        PATH="/opt/local/libexec/gnubin:$PATH"
        
        {load_venv}
        
        peltak test --allow-empty
        
        {command}
        
    '''.format(load_venv=load_venv, command=pre_push)))

    log.info("Making hooks executable")
    if not context.get('pretend', False):
        os.chmod(conf.proj_path('.git/hooks/pre-commit'), 0o755)
        os.chmod(conf.proj_path('.git/hooks/pre-push'), 0o755)


def push():
    """ Push the current branch to origin.

    This is an equivalent of ``git push -u origin <branch>``. Mainly useful for
    the first push as afterwards ``git push`` is just quicker. Free's you from
    having to manually type the current branch name in the first push.
    """
    branch = git.current_branch().name
    shell.run('git push -u origin {}'.format(branch))


def delete_remote():
    """ Delete the current branch on origin.

    This is an equivalent of ``git push origin :<branch>``. Easy way to quickly
    delete the remote branch without having to type in the branch name.
    """
    branch = git.current_branch().name
    shell.run('git push -u origin {}'.format(branch))



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module: peltak.core.templates
    :synopsis: Jinja2 templates support.

#########################
Script template reference
#########################


Using jinja filters
~~~~~~~~~~~~~~~~~~~

Let's suppose you have a scripts that runs pytest over your project and you want
to pass the verbosity flag down to the pytest command. The verbosity option
can be accessed with ``{{ opts.verbose }}`` but it is an ``int`` so we need
to somehow convert it to the appropriate flag ``-v``, or ``-vv`` etc. Thankfully
peltak already implements a filter to do just that, it's called ``count_flag``
and will convert a given number **N** to a flag that has the given letter appear
**N** times. Here's a quick example.


.. code-block:: yaml

    scripts:
        test:
            about: Run tests with pytest
            command: |
                pytest {{ opts.verbose | count_flag('v') }} {{ conf.src_dir }}

This will result in the following command being invoked:

.. code-block:: bash

    peltak run test
    # will result in: pytest src

    peltak run test -v
    # will result in: pytest src -v

    peltak run test -vv
    # will result in: pytest src -vv

Template context
================

Peltak will inject the following context into the template.

================== ================================================================
 Name               Description
------------------ ----------------------------------------------------------------
 ``conf``           | The entire configuration object. This is the dictionary
                    | representation of `pelconf.yaml` and makes it easy to acess
                    | any global configuration values from within a script.
 ``opts``           | Script command line options. This will contains all command
                    | line options the script was called with.
 ``script``         | The script configuration as read from `pelconf.yaml`.
                    | This will only contain the configuration for the currently
                    | running script, not the entire **scripts:** section.
 ``ctx``            | Current runtime context. This is a value store that exists
                    | only when peltak is running and is recreated on every run.
                    | This is a way to share runtime information between commands
                    | (things like **verbosity** or **pretend**).
 ``proj_path``      | A helper function. Given any project relative path (relative
                    | to `pelconf.yaml`) it will convert it to an absolute path.
================== ================================================================

On top of all the values in the context, you can also use
`/reference/script_filters`.


Dev Reference
=============

.. autoclass:: peltak.core.templates.Engine
    :members:

"""
from .engine import Engine  # noqa: F401



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" CLI Commands

The commands should be split into separate py modules, one with the command
click declaration and a separate implementation module. This is to reduce the number
of imports required when loading the CLI for autocompletion. It doesn't change much
for one command module, but makes a big difference when applied to all commands
used by the given peltak configuration. This way the commands module contain only code
used by click to generate the completion and everything else is imported when a given
command is executed.
"""
from typing import Any, Callable, Union

import click

import peltak


AnyFn = Callable[..., Any]


@click.group()
@click.version_option(version=peltak.__version__, message='%(version)s')
def root_cli() -> None:
    """

    To get help for a specific command:

       \033[1m peltak <command> --help\033[0m

    Examples:

       \033[1m peltak lint --help\033[0m

       \033[1m peltak version bump --help\033[0m

       \033[1m peltak release upload --help\033[0m

    """
    pass


def pretend_option(fn: AnyFn) -> AnyFn:
    """ Decorator to add a --pretend option to any click command.

    The value won't be passed down to the command, but rather handled in the
    callback. The value will be accessible through `peltak.core.context` under
    'pretend' if the command needs it. To get the current value you can do:

        >>> from peltak.commands import click, root_cli
        >>> from peltak.core import context
        >>>
        >>> @root_cli.command('my-command')
        ... @pretend_option
        ... def my_command():
        ...     pretend = context.get('pretend', False)

    This value will be accessible from anywhere in the code.
    """

    def set_pretend(     # pylint: disable=missing-docstring
        ctx: click.Context,
        param: Union[click.Option, click.Parameter],
        value: Any
    ) -> Any:
        from peltak.core import context, shell

        context.set('pretend', value or False)
        if value:
            shell.cprint('<90>{}', _pretend_msg())

    return click.option(
        '--pretend',
        is_flag=True,
        help=("Do not actually do anything, just print shell commands that "
              "would be executed."),
        expose_value=False,
        callback=set_pretend
    )(fn)


def _pretend_msg():
    from peltak.core import util
    msg = '''
        ┌─────────────────────────────────────────────────────────────────┐
        │                     Running in pretend mode.                    │
        ├─────────────────────────────────────────────────────────────────┤
        │ All commands that you will see printed in gray are not actually │
        │ executed but printed out instead.                               │
        │                                                                 │
        │ Those would be executed if --pretend option was not specified.  │
        │                                                                 │
        │ You can still add verbosity to some commands using the          │
        │ -v/--verbose option.                                            │
        └─────────────────────────────────────────────────────────────────┘
    '''

    if hasattr(msg, 'decode'):  # py2
        msg = msg.decode('utf-8')

    try:
        # We do empty format() so that it forces unicode errors to happen here
        # and not when it's used/printed.
        return '{}'.format(util.remove_indent(msg))
    except Exception:
        # Pretty printing the frame is not supported - fallback to raw ASCII.
        return util.remove_indent('''
        +-----------------------------------------------------------------+
        |                     Running in pretend mode.                    |
        +-----------------------------------------------------------------+
        | All commands that you will see printed in gray are not actually |
        | executed but printed out instead.                               |
        |                                                                 |
        | Those would be executed if --pretend option was not specified.  |
        |                                                                 |
        | You can still add verbosity to some commands using the          |
        | -v/--verbose option.                                            |
        +-----------------------------------------------------------------+
    ''')


def verbose_option(fn):
    """ Decorator to add a --verbose option to any click command.

    The value won't be passed down to the command, but rather handled in the
    callback. The value will be accessible through `peltak.core.context` under
    'verbose' if the command needs it. To get the current value you can do:

        >>> from peltak.core import context
        >>>
        >>> pretend = context.get('verbose', False)

    This value will be accessible from anywhere in the code.
    """

    def set_verbose(    # pylint: disable=missing-docstring
        ctx: click.Context,
        param: Union[click.Option, click.Parameter],
        value: Any
    ) -> Any:
        from peltak.core import context
        context.set('verbose', value or 0)

    return click.option(
        '-v', '--verbose',
        expose_value=False,
        count=True,
        callback=set_verbose,
        help="Be verbose. Can specify multiple times for more verbosity.",
    )(fn)



from typing import List, Optional

from .. import conf, shell, util
from .types import BranchDetails, CommitDetails


@util.cached_result()
def current_branch() -> BranchDetails:
    """ Return the BranchDetails for the current branch.

    Return:
        BranchDetails: The details of the current branch.
    """
    cmd = 'git symbolic-ref --short HEAD'
    branch_name = shell.run(
        cmd,
        capture=True,
        never_pretend=True
    ).stdout.strip()

    return BranchDetails.parse(branch_name)


@util.cached_result()
def latest_commit() -> CommitDetails:
    """ Return details for the latest commit.

    Returns:
        CommitDetails: The `CommitDetails` instance for the latest commit on the
        current branch.
    """
    return CommitDetails.get()


def commit_details(sha1: str = '') -> CommitDetails:
    """ Return details about a given commit.

    Args:
        sha1 (str):
            The sha1 of the commit to query. If not given, it will return the
            details for the latest commit.

    Returns:
        CommitDetails: A named tuple ``(name, email)`` with the commit author
        details.
    """
    return CommitDetails.get(sha1)


def commit_branches(sha1: str) -> List[str]:
    """ Get the name of the branches that this commit belongs to. """
    cmd = 'git branch --contains {}'.format(sha1)
    return shell.run(
        cmd,
        capture=True,
        never_pretend=True
    ).stdout.strip().split()


@util.cached_result()
def guess_base_branch() -> Optional[str]:
    """ Try to guess the base branch for the current branch.

    Do not trust this guess. git makes it pretty much impossible to guess
    the base branch reliably so this function implements few heuristics that
    will work on most common use cases but anything a bit crazy will probably
    trip this function.

    Returns:
        Optional[str]: The name of the base branch for the current branch if
        guessable or **None** if can't guess.
    """
    my_branch = current_branch(refresh=True).name

    curr = latest_commit()
    if len(curr.branches) > 1:
        # We're possibly at the beginning of the new branch (currently both
        # on base and new branch).
        other = [x for x in curr.branches if x != my_branch]
        if len(other) == 1:
            return other[0]
        return None
    else:
        # We're on one branch
        parent = curr

        while parent and my_branch in parent.branches:
            curr = parent

            if len(curr.branches) > 1:
                other = [x for x in curr.branches if x != my_branch]
                if len(other) == 1:
                    return other[0]
                return None

            parents = [p for p in curr.parents if my_branch in p.branches]
            num_parents = len(parents)

            if num_parents > 2:
                # More than two parent, give up
                return None
            if num_parents == 2:
                # This is a merge commit.
                for p in parents:
                    if p.branches == [my_branch]:
                        parent = p
                        break
            elif num_parents == 1:
                parent = parents[0]
            elif num_parents == 0:
                parent = None

        return None


@util.cached_result()
def branches() -> List[str]:
    """ Return a list of branches in the current repo.

    Returns:
        list[str]: A list of branches in the current repo.
    """
    out = shell.run(
        'git branch',
        capture=True,
        never_pretend=True
    ).stdout.strip()
    return [x.strip('* \t\n') for x in out.splitlines()]


def verify_branch(branch_name: str) -> bool:
    """ Verify if the given branch exists.

    Args:
        branch_name (str):
            The name of the branch to check.

    Returns:
        bool: **True** if a branch with name *branch_name* exits, **False**
        otherwise.
    """
    try:
        shell.run(
            'git rev-parse --verify {}'.format(branch_name),
            never_pretend=True
        )
        return True
    except IOError:
        return False


@util.cached_result()
def protected_branches() -> List[str]:
    """ Return branches protected by deletion.

    By default those are master and devel branches as configured in pelconf.

    Returns:
        list[str]: Names of important branches that should not be deleted.
    """
    master = conf.get('git.master_branch', 'master')
    develop = conf.get('git.devel_branch', 'develop')
    return conf.get('git.protected_branches', (master, develop))



# Copyright 2017-2021 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from typing import List

from .. import conf, shell, util


@util.cached_result()
def untracked() -> List[str]:
    """ Return a list of untracked files in the project repository.

    Returns:
        list[str]: The list of files not tracked by project git repo.
    """
    with conf.within_proj_dir():
        status = shell.run(
            'git status --porcelain',
            capture=True,
            never_pretend=True
        ).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status.startswith('?? '):
                results.append(file_status[3:].strip())

        return results


@util.cached_result()
def unstaged() -> List[str]:
    """ Return a list of unstaged files in the project repository.

    Returns:
        list[str]: The list of files not tracked by project git repo.
    """
    with conf.within_proj_dir():
        status = shell.run(
            'git status --porcelain',
            capture=True,
            never_pretend=True
        ).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status.strip() and file_status[0] == ' ':
                results.append(file_status[3:].strip())

        return results


@util.cached_result()
def staged() -> List[str]:
    """ Return a list of project files staged for commit.

    Returns:
        list[str]: The list of project files staged for commit.
    """
    with conf.within_proj_dir():
        status = shell.run(
            'git status --porcelain',
            capture=True,
            never_pretend=True
        ).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status and file_status[0] in ('A', 'M', 'D'):
                results.append(file_status[3:].strip())

        return results



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import re
from collections import OrderedDict
from os.path import exists
from typing import Optional

from peltak.core import fs, util


RE_PY_VERSION = re.compile(
    r'__version__\s*=\s*["\']'
    r'(?P<version>\d+(\.\d+(\.\d+)?)?)'
    r'["\']'
)
# major.minor[.patch[-build]]
RE_VERSION = re.compile(
    r'^'
    r'(?P<major>\d+)'
    r'(\.(?P<minor>\d+)'
    r'(\.(?P<patch>\d+))?)?'
    r'$'
)


class InvalidVersionFile(RuntimeError):
    pass


def is_valid(version_str: str) -> bool:
    """ Check if the given string is a version string

    Args:
        version_str (str):
            A string to check

    Returns:
        bool: **True** if the given string is a version.
    """
    return bool(version_str and RE_VERSION.match(version_str))


class VersionFile(object):
    """ Base class for version storage.

    A version storage is a way to store the project version. Different projects
    will have different ways of storing the version. The simplest case is a
    text file that holds just the version number (probably called VERSION).
    Python projects can also use a version stored as ``__version__`` variable
    inside the project package/module. A Node.js project on the other hand will
    probably keep the version in ``package.json``. All of the above strategies
    can be (and are) implemented through subclassing this class.

    @see `PyVersionFile`, `RawVersionFile`, `NodeVersionFile`
    """
    def __init__(self, path: str) -> None:
        self.path = path
        if not exists(path):
            raise ValueError(f"Version file '{path}' does not exist.")

    def read(self) -> Optional[str]:
        """ Read the current project version.

        All subclasses must implement this method.

        Raises:
            NotImplementedError:
                If the subclass does nto implement the read() method.
        """
        raise NotImplementedError(f"{self.__class__.__name__} must implement .read()")

    def write(self, version: str):
        """ Save the given version as the current project version.

        All subclasses must implement this method.

        Raises:
            NotImplementedError:
                If the subclass does nto implement the write() method.
        """
        raise NotImplementedError(f"{self.__class__.__name__} must implement .write()")


class PyVersionFile(VersionFile):
    """ Store project version in one of the py module/package files. """
    def read(self) -> Optional[str]:
        """ Read the project version from .py file.

        This will regex search in the file for a
        ``__version__ = VERSION_STRING`` and read the version string.
        """
        with open(self.path) as fp:
            content = fp.read()
            m = RE_PY_VERSION.search(content)
            if not m:
                return None
            else:
                return m.group('version')

    def write(self, version: str):
        """ Write the project version to .py file.

        This will regex search in the file for a
        ``__version__ = VERSION_STRING`` and substitute the version string
        for the new version.
        """
        with open(self.path) as fp:
            content = fp.read()

        ver_statement = "__version__ = '{}'".format(version)
        new_content = RE_PY_VERSION.sub(ver_statement, content)
        fs.write_file(self.path, new_content)


class RawVersionFile(VersionFile):
    """ Store project version as a simple value in a text file. """
    def read(self) -> Optional[str]:
        """ Read the project version from .py file.

        This will regex search in the file for a
        ``__version__ = VERSION_STRING`` and read the version string.
        """
        with open(self.path) as fp:
            version = fp.read().strip()

            if is_valid(version):
                return version

            return None

    def write(self, version: str):
        fs.write_file(self.path, version)


class NodeVersionFile(VersionFile):
    """ Store project version in package.json. """
    def read(self) -> Optional[str]:
        with open(self.path) as fp:
            config = json.load(fp)
            return config.get('version')

    def write(self, version: str):
        with open(self.path, 'r') as fp:
            config = json.load(fp, object_pairs_hook=OrderedDict)

        config['version'] = version

        fs.write_file(self.path, json.dumps(config, indent=2) + '\n')


class PoetryVersionFile(VersionFile):
    """ Store project version in package.json. """
    def read(self) -> Optional[str]:
        config = util.toml_load(self.path)
        return util.get_from_dict(config, 'tool.poetry.version', None)

    def write(self, version: str):
        config = util.toml_load(self.path)
        config['tool']['poetry']['version'] = version

        fs.write_file(self.path, util.toml_dump(config))


# path test -> VersionFile implementation
VERSION_FILE_TYPES = [
    (lambda p: p.endswith('.py'), PyVersionFile),
    (lambda p: p.endswith('package.json'), NodeVersionFile),
    (lambda p: p.endswith('pyproject.toml'), PoetryVersionFile),
    (lambda p: True, RawVersionFile),   # default to Raw version file.
]


def load_version_file(path: str) -> VersionFile:
    for test, cls in VERSION_FILE_TYPES:
        if test(path):
            return cls(path)

    raise InvalidVersionFile()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cliform',
 'peltak',
 'peltak.commands',
 'peltak.core',
 'peltak.core.git',
 'peltak.core.scripts',
 'peltak.core.templates',
 'peltak.core.versioning',
 'peltak.testing']

package_data = \
{'': ['*'], 'peltak': ['templates/*']}

install_requires = \
['PyYAML>=6.0',
 'click>=7.0,<8.0',
 'jinja2>=2.12',
 'pygments>=2.7.4',
 'tomlkit>=0.7.0']

entry_points = \
{'console_scripts': ['peltak = peltak.main:root_cli']}

setup_kwargs = {
    'name': 'peltak',
    'version': '0.28.5',
    'description': 'A command line tool to help manage a project',
    'long_description': '######\npeltak\n######\n\n.. readme_about_start\n\n**peltak**\'s goal is to simplify day-to-day dev tasks that we all, as developers,\nhave to do. It does not try to be a build system or to replace any of the tools\nyou would normally use when developing project like test runners, linters, etc.\nThink of it rather as a replacement for the shell scripts that usually wrap\nthose tools. It actually started years ago as a collection of those shell\nscripts that with time evolved into a very flexible, extensible and easy to use\ncommand line app.\n\n.. readme_about_end\n\nUseful links\n============\n\n- `Documentation <https://novopl.github.io/peltak>`_\n    - `Installation <https://novopl.github.io/peltak/docs/html/guides/installation.html>`_\n    - `Quickstart <https://novopl.github.io/peltak/docs/html/guides/quickstart.html>`_\n    - `Dev Docs <https://novopl.github.io/peltak/docs/html/dev/_index.html>`_\n- `Source Code <https://github.com/novopl/peltak>`_\n- `CI Builds <https://circleci.com/gh/novopl/peltak>`_\n\n\nInstallation\n============\n\n.. readme_installation_start\n\n.. code-block:: shell\n\n    $ pip install peltak\n\nEnabling auto-completion\n------------------------\n**peltak** has a great auto-completion thanks to the underlying click library.\nThe steps to enable it vary slightly depending on what shell you are using\n\n**Bash users**\n\n    Either run this command or to make the change permanent add it to your\n    ``~/.bashrc``:\n\n    .. code-block:: shell\n\n        eval "$(_PELTAK_COMPLETE=source peltak)"\n\n**ZSH users**\n    Either run this command or to make the change permanent add it to your\n    ``~/.zshrc``:\n\n    .. code-block:: shell\n\n        eval "$(_PELTAK_COMPLETE=source_zsh peltak)"\n\n.. readme_installation_end\n',
    'author': 'Mateusz Klos',
    'author_email': 'novopl@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://novopl.github.com/peltak',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0',
}


setup(**setup_kwargs)



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
##################
``peltak version``
##################

Commands for managing the managed project version.
"""
from typing import Optional

from peltak.commands import click, pretend_option, root_cli


@root_cli.group('version', invoke_without_command=True)
@click.option('--porcelain', is_flag=True)
@click.pass_context
def version_cli(ctx: click.Context, porcelain: bool) -> None:
    """ Show project version. Has sub commands.

    For this command to work you must specify where the project version is
    stored. You can do that with version.file conf variable. peltak supports
    multiple ways to store the project version. Right now you can store it in a
    python file using built-in __version__ variable. You can use node.js
    package.json and keep the version there or you can just use a plain text
    file that just holds the raw project version. The appropriate storage is
    guessed based on the file type and name.

    Example Configuration::

        version:
            file: 'src/mypackage/__init__.py'

    Examples:

        \b
        $ peltak version                        # Pretty print current version
        $ peltak version --porcelain            # Print version as raw string
        $ peltak version bump patch             # Bump patch version component
        $ peltak version bump minor             # Bump minor version component
        $ peltak version bump major             # Bump major version component
        $ peltak version bump release           # same as version bump patch
        $ peltak version bump --exact=1.2.1     # Set project version to 1.2.1

    """
    if ctx.invoked_subcommand:
        return

    from peltak.core import log, versioning

    current = versioning.current()

    if porcelain:
        print(current)
    else:
        log.info("Version: <35>{}".format(current))


@version_cli.command('bump')
@click.argument(
    'component',
    type=click.Choice(['major', 'minor', 'patch']),
    required=False,
    default='patch'
)
@click.option('--exact', type=str)
@pretend_option
def bump_version(component: str = 'patch', exact: Optional[str] = None):
    """ Bump current project version without committing anything.

    No tags are created either.

    Examples:

        \b
        $ peltak version bump patch             # Bump patch version component
        $ peltak version bump minor             # Bump minor version component
        $ peltak version bump major             # Bump major version component
        $ peltak version bump release           # same as version bump patch
        $ peltak version bump --exact=1.2.1     # Set project version to 1.2.1

    """
    from peltak.core import log, versioning

    old_ver, new_ver = versioning.bump(component, exact)

    log.info("Project version bumped")
    log.info("  old version: <35>{}".format(old_ver))
    log.info("  new version: <35>{}".format(new_ver))



""" Types and classes used across **peltak** codebase. """
import dataclasses
from typing import Any, Callable, Dict, List, Type, Union


AnyFn = Callable[..., Any]
YamlConf = Dict[str, Any]
CliOptions = Dict[str, Any]
PlainDict = Dict[str, Any]
JsonDict = Union[PlainDict, List[Any]]
Decorator = Callable[[AnyFn], AnyFn]


@dataclasses.dataclass
class FilesCollection:
    """ Configure files passed to the script.

    This section allows you to inject a set of files into the script as the
    ``{{ files }}`` expression. This configuration allows you to flexibly define
    what files are collected by specifying the *paths* of interest and further
    refining the list by *include* and *exclude* which act as a pattern based
    whitelist and blacklist for the paths specified in *paths*.

    On top of that you also have 2 boolean flags: *commit* will collect
    only files staged for commit and *untracked* (``True`` by default will
    include or not files untracked by git).
    """
    paths: List[str]
    include: List[str] = dataclasses.field(default_factory=list)
    exclude: List[str] = dataclasses.field(default_factory=list)
    only_staged: bool = False
    untracked: bool = True
    use_gitignore: bool = True

    @classmethod
    def from_config(
        cls: Type['FilesCollection'],
        files_conf: YamlConf,
    ) -> 'FilesCollection':
        """ Load from config dict """
        paths = files_conf.get('paths')
        fields = {f.name: f for f in dataclasses.fields(cls)}
        include = files_conf.get(
            'include',
            fields['include'].default_factory()  # type: ignore
        )
        exclude = files_conf.get(
            'exclude',
            fields['exclude'].default_factory()  # type: ignore
        )

        if not paths:
            raise ValueError("You must define the paths when using script files")

        # A string value is the same as one element array.
        paths = [paths] if isinstance(paths, str) else paths
        include = [include] if isinstance(include, str) else include
        exclude = [exclude] if isinstance(exclude, str) else exclude

        return cls(
            paths=paths,
            include=include,
            exclude=exclude,
            only_staged=files_conf.get('only_staged', fields['only_staged'].default),
            untracked=files_conf.get('untracked', fields['untracked'].default),
            use_gitignore=files_conf.get(
                'use_gitignore', fields['use_gitignore'].default,
            ),
        )

    def whitelist(self) -> List[str]:
        """ Return a full whitelist for use with `fs.filtered_walk()` """
        from peltak.core import fs  # avoid circular dependency.
        from peltak.core import git

        if self.only_staged:
            # Only include committed files if commit only is true.
            include = [
                '*' + f for f in git.staged()
                if not self.include or fs.match_globs(f, self.include)
            ]

        else:
            include = list(self.include)

        return include

    def blacklist(self) -> List[str]:
        """ Return a full blacklist for use with `fs.filtered_walk()` """
        from peltak.core import git

        exclude = list(self.exclude)

        # prepare
        if self.use_gitignore:
            exclude += git.ignore()

        if not self.untracked:
            exclude += git.untracked()

        return exclude



import os
import re
from pathlib import Path
from typing import Any, Dict, Iterator, List, NamedTuple, Tuple, cast

from peltak.commands import root_cli
from peltak.core import conf, exc, util

from . import types


class ScriptId(NamedTuple):
    name: str
    path: str


class NoScript(exc.PeltakError):
    msg = "No Script"


CliGroups = Dict[str, Any]
ScriptsMap = Dict[ScriptId, types.Script]


def register_scripts_from(scripts_dir: Path) -> None:
    """ Parse script files and build the ClI for it. """
    cli_groups: Dict[str, Any] = {}
    if scripts_dir.exists() and scripts_dir.is_dir():
        # Silently return if the scripts directory does not exist. They are not
        # required and the completion should not brake so we can't raise
        # any exceptions or print anything to stdout/stderr.
        script_files = list(_iter_script_files(scripts_dir))
        scripts = _process_script_files(scripts_dir, script_files)
        cli_groups = generate_cli_groups(scripts)

        for script_id, script in scripts.items():
            cli_group = cli_groups[script_id.path]
            script.register(cli_group)

    # Load the scripts defined in the old way. This is now deprecated and will
    # be removed before v1.0
    # TODO: Remove old scripts implementation.
    old_scripts = conf.get('scripts', {})
    if 'run' in cli_groups:
        run_cli = cli_groups['run']
    else:
        run_cli = root_cli.group('run')(lambda: None)

    for name, script_conf in old_scripts.items():
        script = types.Script.from_config(name, script_conf)
        script.register(root_cli if script.root_cli else run_cli)


def generate_cli_groups(scripts: ScriptsMap) -> Dict[str, Any]:
    results: Dict[str, Any] = {'': root_cli}
    unique_paths = sorted(frozenset(x.path for x in scripts.keys() if x.path))

    for path in unique_paths:
        results[path] = _generate_cli_group(results, path)

    return results


def _iter_script_files(scripts_dir: Path) -> Iterator[Path]:
    results: List[Path] = []

    for dir_item in scripts_dir.iterdir():
        if dir_item.is_dir():
            yield from _iter_script_files(dir_item)
        # else:
        #     yield dir_item
        elif dir_item.name.endswith('.sh'):
            # We only support shell scripts this way. The user can also keep python
            # scripts in his scripts dir and those should import `root_cli` and
            # be registered via ``pelconf.yaml``
            # TODO: Autoload python scripts found inside `scripts_dir`.
            yield dir_item
        elif dir_item.name.endswith('.py'):
            # Auto load python scripts in scripts_dir
            conf.py_import(dir_item.stem)

    return results


def _process_script_files(scripts_dir: Path, script_files: List[Path]) -> ScriptsMap:
    results: ScriptsMap = {}

    for script_file in sorted(script_files):
        rel_path = script_file.relative_to(scripts_dir)
        script_id = _gen_script_id(rel_path)
        try:
            results[script_id] = _parse_script(script_id, script_file)
        except NoScript:
            pass

    return results


def _parse_script(script_id: ScriptId, script_path: Path) -> types.Script:
    """ Parse a script file.

    The file starts wit the file header.
    """
    re_header = re.compile(r'^# (?P<content>.*)')
    header_lines: List[str] = []
    source_lines: List[str] = []
    header_ended = False

    with script_path.open() as fp:
        lines = fp.readlines()
        if not lines:
            raise ValueError(f"Empty file {script_path}")

        if lines[0].startswith('#!'):
            lines = lines[1:]

        for line in lines:
            # Only the top lines are considered header, once it ends we won't
            # search for it anymore
            if not header_ended:
                m = re_header.match(line)
                if m:
                    content = m.group('content')
                    # An empty comment line also ends the header
                    if content.strip():
                        header_lines.append(content)
                    else:
                        header_ended = True
                    continue
                else:
                    # First line that is not a comment means it's not a header anymore
                    header_ended = True

            source_lines.append(line.rstrip())

    header_yaml = '\n'.join(header_lines)

    if not header_yaml:
        raise NoScript(f"peltak header is missing from file: {script_path}")

    header: Dict[str, Any] = cast(Dict[str, Any], util.yaml_load(header_yaml))
    if not (len(header) == 1 and 'peltak' in header):
        raise NoScript("Script header can only have a single root object 'peltak'")

    return types.Script.from_config(
        name=script_id.name,
        script_conf={
            **header['peltak'],
            'command': '\n'.join(source_lines).strip(),
        },
    )


def _gen_script_id(rel_path: Path) -> ScriptId:
    path = ' '.join(str(rel_path).split(os.sep)[:-1])
    name = rel_path.name.split('.')[0]
    return ScriptId(name, path)


def _generate_cli_group(groups: Dict[str, Any], group_path: str):
    name, parent = _parse_group_path(group_path)
    parent_cli = root_cli

    if parent:
        if parent in groups:
            parent_cli = groups[parent]
        else:
            parent_cli = _generate_cli_group(groups, parent)
            groups[parent] = parent_cli

    return parent_cli.group(name)(lambda: None)


def _parse_group_path(group_path: str) -> Tuple[str, str]:
    parts = group_path.rsplit(maxsplit=1)
    name = parts[-1]
    parent = parts[0] if len(parts) == 2 else ''
    return name, parent



""" A place to store all mocks specific to peltak.

Easier to track them down, when you need one.
"""
from typing import Optional

from peltak.core.shell import ExecResult


def mock_result(
    stdout: Optional[str] = None,
    retcode: Optional[int] = None,
    stderr: Optional[str] = None,
    cmd: Optional[str] = None
) -> ExecResult:
    """ Helper for creating ExecResults for tests.

    Args:
        stdout (str):
        retcode (int):
        stderr (str):
        cmd (str):

    Returns:
        ExecResult:
    """
    return ExecResult(
        cmd or '',
        retcode or 0,
        stdout or '',
        stderr or '',
        retcode == 0,
        retcode != 0,
    )



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Types and classes used by ``peltak.extra.scripts``. """
import dataclasses
from typing import Any, Callable, Dict, List, Optional, Type, cast

from peltak.commands import click, pretend_option, verbose_option
from peltak.core import types


AnyFn = Callable[..., Any]
YamlConf = Dict[str, Any]
CliOptions = Dict[str, Any]


@dataclasses.dataclass
class ScriptOption:
    """ Describes the scripts command line option. """
    name: List[str]
    default: Optional[Any] = None
    about: str = ''
    is_flag: bool = False
    count: bool = False
    type: Type = str

    @classmethod
    def from_config(cls: Type['ScriptOption'], option_conf: YamlConf) -> 'ScriptOption':
        """ Load script option from configuration in pelconf.yaml """
        fields = {f.name: f for f in dataclasses.fields(cls)}
        name = option_conf.get('name')

        if not name:
            raise ValueError("You must define the name of the option")

        if isinstance(name, str):
            # Support passing name: ['--opt'] or name: '--opt'
            name = [name]

        # Convert string type to an initializer for that type.
        if 'type' not in option_conf:
            opt_type = fields['type'].default
        else:
            opt_type = {
                'str': str,
                'int': int,
                'float': float,
            }.get(option_conf['type'])

            if not opt_type:
                raise ValueError("Unsupported {} option type {}".format(
                    name, option_conf['type']
                ))

        return cls(
            name=name,
            default=option_conf.get('default', fields['default'].default),
            about=option_conf.get('about', fields['about'].default),
            is_flag=option_conf.get('is_flag', fields['is_flag'].default),
            count=option_conf.get('count', fields['count'].default),  # type: ignore
            type=cast(Type, opt_type),
        )


@dataclasses.dataclass
class Script:
    """ Represents a single script defined in pelconf.yaml.

    The important thing is to do a little as possible during creation and only
    do heavy stuff inside the `Script.run()` method. This is because the
    scripts are parsed in a *post-conf-load* hook and that code impacts the
    speed of shell auto completion for peltak command.
    """
    name: str
    command: str
    command_file: Optional[str] = None
    about: str = ''
    root_cli: bool = False
    success_exit_codes: List[int] = dataclasses.field(default_factory=lambda: [0])
    options: List[ScriptOption] = dataclasses.field(default_factory=list)
    files: Optional[types.FilesCollection] = None

    @classmethod
    def from_config(cls, name: str, script_conf: YamlConf) -> 'Script':
        """ Load script from pelfconf.yaml """
        fields = {f.name: f for f in dataclasses.fields(cls)}
        options = script_conf.get(
            'options',
            fields['options'].default_factory(),  # type: ignore
        )
        files = None
        success_exit_codes = script_conf.get(
            'success_exit_codes',
            fields['success_exit_codes'].default_factory(),   # type: ignore
        )

        # Parse 'files' section if it's present
        if 'files' in script_conf:
            files = types.FilesCollection.from_config(script_conf['files'])

        if isinstance(success_exit_codes, int):
            success_exit_codes = [success_exit_codes]

        # Cannot have a script without a 'command' or `command_file`.
        if 'command' not in script_conf and 'command_file' not in script_conf:
            raise ValueError(
                "'command' or 'command_file' must be specified "
                "for script '{}'".format(name)
            )

        return cls(
            name=name,
            command=script_conf.get('command', ''),
            command_file=script_conf.get('command_file', ''),
            about=script_conf.get('about', fields['about'].default),
            root_cli=script_conf.get('root_cli', fields['root_cli'].default),
            success_exit_codes=success_exit_codes,
            options=[ScriptOption.from_config(opt_conf) for opt_conf in options],
            files=files,
        )

    def register(self, cli_group: Any):
        """ Register the script with click.

        *cli_group* is the result of using the ``@click.group()`` decorator.
        """
        @verbose_option
        @pretend_option
        @click.pass_context
        def script_command(ctx, **options):  # pylint: disable=missing-docstring
            from .logic import run_script  # nocov
            run_script(self, options)   # nocov

        script_command.__doc__ = self.about

        # Add all option definitions to the generated click command.
        for option in self.options:
            script_command = self._add_option(script_command, option)

        cli_group.command(self.name)(script_command)

    def _add_option(self, cmd_fn: AnyFn, option: ScriptOption) -> AnyFn:
        return click.option(
            *option.name,
            is_flag=option.is_flag,
            default=option.default,
            help=option.about,
            count=option.count,
            type=option.type
        )(cmd_fn)

    @property
    def header(self) -> Dict[str, Any]:
        result = dataclasses.asdict(self)
        del result['command']
        return result
